// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

package frc.robot.subsystems;

import java.io.File;
import java.util.Optional;
import java.util.function.BooleanSupplier;
import java.util.function.DoubleSupplier;

import org.littletonrobotics.junction.Logger;

import com.ctre.phoenix6.Orchestra;
import com.pathplanner.lib.auto.AutoBuilder;
import com.studica.frc.AHRS;
import com.studica.frc.AHRS.NavXComType;

import edu.wpi.first.math.MathUtil;
import edu.wpi.first.math.VecBuilder;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.math.estimator.SwerveDrivePoseEstimator;
import edu.wpi.first.math.geometry.Pose2d;
import edu.wpi.first.math.geometry.Rotation2d;
import edu.wpi.first.math.geometry.Translation2d;
import edu.wpi.first.math.interpolation.TimeInterpolatableBuffer;
import edu.wpi.first.math.kinematics.ChassisSpeeds;
import edu.wpi.first.math.kinematics.SwerveDriveKinematics;
import edu.wpi.first.math.kinematics.SwerveModulePosition;
import edu.wpi.first.math.kinematics.SwerveModuleState;
import edu.wpi.first.math.util.Units;
import edu.wpi.first.wpilibj.DriverStation;
import edu.wpi.first.wpilibj.DriverStation.Alliance;
import edu.wpi.first.wpilibj.Filesystem;
import edu.wpi.first.wpilibj.Timer;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.AutoConstants;
import frc.robot.constants.DrivetrainConstants;
import frc.robot.constants.OIConstants;
import frc.robot.constants.VisionConstants;
import edu.wpi.first.math.geometry.Pose3d;

public class Drivetrain extends SubsystemBase {
    // Create MAXSwerveModules
    protected MAXSwerveModule m_frontLeft;
    protected MAXSwerveModule m_frontRight;
    protected MAXSwerveModule m_rearLeft;
    protected MAXSwerveModule m_rearRight;

    // The gyro sensor
    private AHRS gyro;

    // Odometry class for tracking robot pose
    private SwerveDrivePoseEstimator m_estimator;
    
    private TimeInterpolatableBuffer<Double> gyroBuffer = TimeInterpolatableBuffer.createDoubleBuffer(2.0);

    public Orchestra m_orchestra = new Orchestra();
    private Timer musicTimer = new Timer();

    private PIDController pidHeading;
    private PIDController pidTranslationX;
    private PIDController pidTranslationY;

    public Vision vision;

    public Pose2d orangePose2d;
    public Pose2d blackPose2d;

    /** Creates a new DriveSubsystem. */
    public Drivetrain() {
        m_frontLeft = new MAXSwerveModule(
            DrivetrainConstants.kFrontLeftDrivingCanId,
            DrivetrainConstants.kFrontLeftTurningCanId,
            DrivetrainConstants.kFrontLeftChassisAngularOffset
        );

        m_frontRight = new MAXSwerveModule(
            DrivetrainConstants.kFrontRightDrivingCanId,
            DrivetrainConstants.kFrontRightTurningCanId,
            DrivetrainConstants.kFrontRightChassisAngularOffset
        );

        m_rearLeft = new MAXSwerveModule(
            DrivetrainConstants.kRearLeftDrivingCanId,
            DrivetrainConstants.kRearLeftTurningCanId,
            DrivetrainConstants.kBackLeftChassisAngularOffset
        );

        m_rearRight = new MAXSwerveModule(
            DrivetrainConstants.kRearRightDrivingCanId,
            DrivetrainConstants.kRearRightTurningCanId,
            DrivetrainConstants.kBackRightChassisAngularOffset
        );

        gyro = new AHRS(NavXComType.kMXP_SPI);
        gyro.reset();

        m_estimator = new SwerveDrivePoseEstimator(
            DrivetrainConstants.kDriveKinematics,
            Rotation2d.fromDegrees(getGyroValue()),
            new SwerveModulePosition[] {
                m_frontLeft.getPosition(),
                m_frontRight.getPosition(),
                m_rearLeft.getPosition(),
                m_rearRight.getPosition()
            },
            new Pose2d(),
            VecBuilder.fill(0.05, 0.05, Units.degreesToRadians(5)),
            VecBuilder.fill(1, 1, Units.degreesToRadians(360))
        );

        pidHeading = new PIDController(DrivetrainConstants.kHeadingP,0,0);
        pidHeading.enableContinuousInput(-180, 180);

        pidTranslationX = new PIDController(DrivetrainConstants.kXTranslationP,0,0);
        pidTranslationX.setTolerance(Units.inchesToMeters(0.5));
        pidTranslationY = new PIDController(DrivetrainConstants.kYTranslationP,0,0);
        pidTranslationY.setTolerance(Units.inchesToMeters(0.5));


        AutoBuilder.configure(
            this::getPose, 
            this::resetOdometry, 
            this::getCurrentChassisSpeeds, 
            this::driveWithChassisSpeeds, 
            AutoConstants.kPPDriveController, 
            AutoConstants.kRobotConfig,
            () -> {
                Optional<DriverStation.Alliance> alliance = DriverStation.getAlliance();
                    if (alliance.isPresent()) {
                        return alliance.get() == DriverStation.Alliance.Red;
                    }
                    return false;
            }, 
            this
        );

        m_orchestra.loadMusic(Filesystem.getDeployDirectory()
        .toPath()
        .resolve("Orchestra" + File.separator + "doomE1M1.chrp")
        .toString());

        // Add a single device to the orchestra
        m_orchestra.addInstrument(m_frontLeft.getDrivingMotor(), 0);
        m_orchestra.addInstrument(m_frontRight.getDrivingMotor(), 1);
        m_orchestra.addInstrument(m_rearLeft.getDrivingMotor(), 2);
        m_orchestra.addInstrument(m_rearRight.getDrivingMotor(), 3);

        m_orchestra.play();
        musicTimer.reset();
        musicTimer.start();

        vision = new Vision();
        orangePose2d = new Pose2d();
        blackPose2d = new Pose2d();
    }

    @Override
    public void periodic() {
        // Update the odometry in the periodic block
        m_estimator.update(
            Rotation2d.fromDegrees(getGyroValue()),
            new SwerveModulePosition[] {
                m_frontLeft.getPosition(),
                m_frontRight.getPosition(),
                m_rearLeft.getPosition(),
                m_rearRight.getPosition()
        });

        gyroBuffer.addSample(Timer.getFPGATimestamp(), getGyroValue());

        m_estimator.setVisionMeasurementStdDevs(VecBuilder.fill(1, 1, Units.degreesToRadians(360)));

        if(vision.getOrangeTagDetected()){
            if(vision.getOrangeDist() < 60){
                m_estimator.setVisionMeasurementStdDevs(VecBuilder.fill(0.4, 0.4, Units.degreesToRadians(360)));
            }

            // if the detected tags match your alliances reef tags use their pose estimates  
            if(vision.getOrangeClosestTag() >= 6 && vision.getOrangeClosestTag() <= 11 && DriverStation.getAlliance().get().equals(Alliance.Red) && vision.getOrangeTagDetected()){
                orangePose2d = vision.getOrangeGlobalPose(gyroBuffer);
                m_estimator.addVisionMeasurement(orangePose2d, vision.getOrangeTimeStamp());

            }else if(vision.getOrangeClosestTag() >= 17 && vision.getOrangeClosestTag() <= 22 && DriverStation.getAlliance().get().equals(Alliance.Blue) && vision.getOrangeTagDetected()){
                orangePose2d = vision.getOrangeGlobalPose(gyroBuffer);
                m_estimator.addVisionMeasurement(orangePose2d, vision.getOrangeTimeStamp());
            }
            Logger.recordOutput("orange pose", new Pose3d());
        }

        if(vision.getBlackTagDetected()){
            if(vision.getBlackDist() < 60){
                m_estimator.setVisionMeasurementStdDevs(VecBuilder.fill(0.4, 0.4, Units.degreesToRadians(360)));
            }

            if(vision.getBlackClosestTag() >= 6 && vision.getBlackClosestTag() <= 11 && DriverStation.getAlliance().get().equals(Alliance.Red) && vision.getBlackTagDetected()){
                blackPose2d = vision.getBlackGlobalPose(gyroBuffer);
                m_estimator.addVisionMeasurement(blackPose2d, vision.getBlackTimeStamp()); 

            }else if(vision.getBlackClosestTag() >= 17 && vision.getBlackClosestTag() <= 22 && DriverStation.getAlliance().get().equals(Alliance.Blue) && vision.getBlackTagDetected()){
                blackPose2d = vision.getBlackGlobalPose(gyroBuffer);
                m_estimator.addVisionMeasurement(blackPose2d, vision.getBlackTimeStamp());
            }
            Logger.recordOutput("black pose", new Pose3d(blackPose2d));
        }
        
        
        Logger.recordOutput("robot pose", new Pose3d(m_estimator.getEstimatedPosition()));
        
        if(musicTimer.get()>10){
            if (m_orchestra.isPlaying()) {
                m_orchestra.stop();
            }
            musicTimer.stop();
            musicTimer.reset();
        }
        
    }

    public ChassisSpeeds getCurrentChassisSpeeds() {
        return DrivetrainConstants.kDriveKinematics.toChassisSpeeds(
            m_frontLeft.getState(),
            m_frontRight.getState(),
            m_rearLeft.getState(),
            m_rearRight.getState()
        );
    }

    public void driveWithChassisSpeeds(ChassisSpeeds speeds) {
        ChassisSpeeds discreteSpeeds = ChassisSpeeds.discretize(speeds, 0.2);
        SwerveModuleState[] newStates = DrivetrainConstants.kDriveKinematics.toSwerveModuleStates(discreteSpeeds);
        SwerveDriveKinematics.desaturateWheelSpeeds(newStates, DrivetrainConstants.kMaxSpeedMetersPerSecond);

        setModuleStates(newStates);
    }

    /**
     * Returns the currently-estimated pose of the robot.
     *
     * @return The pose.
     */
    public Pose2d getPose() {
        return m_estimator.getEstimatedPosition();
    }

    /**
     * Resets the odometry to the specified pose.
     *
     * @param pose The pose to which to set the odometry.
     */
    public void resetOdometry(Pose2d pose) {
        m_estimator.resetPose(
            pose
        );
    }

    public Command drive(DoubleSupplier xSpeed, DoubleSupplier ySpeed, DoubleSupplier rot,
            BooleanSupplier fieldRelative) {
        return run(() -> {
            if(DriverStation.getAlliance().get().equals(Alliance.Blue)){
                drive(
                    -MathUtil.applyDeadband(xSpeed.getAsDouble(), OIConstants.kDriveDeadband),
                    -MathUtil.applyDeadband(ySpeed.getAsDouble(), OIConstants.kDriveDeadband),
                    -MathUtil.applyDeadband(rot.getAsDouble(), OIConstants.kDriveDeadband),
                    fieldRelative.getAsBoolean()
                );
            }else{
                drive(
                    MathUtil.applyDeadband(xSpeed.getAsDouble(), OIConstants.kDriveDeadband),
                    MathUtil.applyDeadband(ySpeed.getAsDouble(), OIConstants.kDriveDeadband),
                    -MathUtil.applyDeadband(rot.getAsDouble(), OIConstants.kDriveDeadband),
                    fieldRelative.getAsBoolean()
            );}
        });
    }

    /**
     * Method to drive the robot using joystick info.
     *
     * @param xSpeed        Speed of the robot in the x direction (forward).
     * @param ySpeed        Speed of the robot in the y direction (sideways).
     * @param rot           Angular rate of the robot.
     * @param fieldRelative Whether the provided x and y speeds are relative to the
     *                      field.
     */
    public void drive(double xSpeed, double ySpeed, double rot, boolean fieldRelative) {
        // Convert the commanded speeds into the correct units for the drivetrain
        double xSpeedDelivered = xSpeed * DrivetrainConstants.kMaxSpeedMetersPerSecond;
        double ySpeedDelivered = ySpeed * DrivetrainConstants.kMaxSpeedMetersPerSecond;
        double rotDelivered = rot * DrivetrainConstants.kMaxAngularSpeed;

        var swerveModuleStates = DrivetrainConstants.kDriveKinematics.toSwerveModuleStates(
            fieldRelative
                ? ChassisSpeeds.fromFieldRelativeSpeeds(xSpeedDelivered, ySpeedDelivered, rotDelivered,
                        Rotation2d.fromDegrees(getGyroValue()))
                : new ChassisSpeeds(xSpeedDelivered, ySpeedDelivered, rotDelivered));
        SwerveDriveKinematics.desaturateWheelSpeeds(
            swerveModuleStates, DrivetrainConstants.kMaxSpeedMetersPerSecond);
        m_frontLeft.setDesiredState(swerveModuleStates[0]);
        m_frontRight.setDesiredState(swerveModuleStates[1]);
        m_rearLeft.setDesiredState(swerveModuleStates[2]);
        m_rearRight.setDesiredState(swerveModuleStates[3]);
    }

    public Command setXCommand() {
        return run(() -> {
            setX();
        });
    }

    public Command goToPose(DoubleSupplier xSetpoint, DoubleSupplier ySetpoint, DoubleSupplier headingSetpoint){
        return run(() -> {
            drive(MathUtil.clamp(pidTranslationX.calculate(m_estimator.getEstimatedPosition().getX(), xSetpoint.getAsDouble()), -0.1, 0.1),
                MathUtil.clamp(pidTranslationY.calculate(m_estimator.getEstimatedPosition().getY(), ySetpoint.getAsDouble()), -0.1, 0.1), 
                pidHeading.calculate(getHeading(), headingSetpoint.getAsDouble()), 
                true);

                Logger.recordOutput("reef setpoint", new Pose3d(new Pose2d(
                    new Translation2d(xSetpoint.getAsDouble(), ySetpoint.getAsDouble()),
                    new Rotation2d(headingSetpoint.getAsDouble()))));
            });
    }

    public int getClosestTag(){

		if(DriverStation.getAlliance().get().equals(DriverStation.Alliance.Blue)){
			int closestTag = 17;
			double closestTagDist = Math.sqrt(Math.pow(getPose().getX()- Units.inchesToMeters(VisionConstants.globalTagCoords[17][0]), 2)
			+ Math.pow(getPose().getY()- Units.inchesToMeters(VisionConstants.globalTagCoords[17][1]), 2));
			
			for(int i = 17; i <= 22; ++i){
				double distance = Math.sqrt(Math.pow(getPose().getX()- Units.inchesToMeters(VisionConstants.globalTagCoords[i][0]), 2)
				+ Math.pow(getPose().getY()- Units.inchesToMeters(VisionConstants.globalTagCoords[i][1]), 2));

				if(distance < closestTagDist){
					closestTag = i;
					closestTagDist = distance;
				}
			}
			return closestTag;
		}else{
			int closestTag = 6;
			double closestTagDist = Math.sqrt(Math.pow(m_estimator.getEstimatedPosition().getX()- Units.inchesToMeters(VisionConstants.globalTagCoords[6][0]), 2)
			+ Math.pow(m_estimator.getEstimatedPosition().getY()- Units.inchesToMeters(VisionConstants.globalTagCoords[6][1]), 2));
			
			for(int i = 6; i <= 11; ++i){
				double distance = Math.sqrt(Math.pow(m_estimator.getEstimatedPosition().getX()- Units.inchesToMeters( VisionConstants.globalTagCoords[i][0]), 2)
				+ Math.pow(m_estimator.getEstimatedPosition().getY()- Units.inchesToMeters(VisionConstants.globalTagCoords[i][1]), 2));

				if(distance < closestTagDist){
					closestTag = i;
					closestTagDist = distance;
				}
			}
			return closestTag;
		}
	}

    /**
     * Sets the wheels into an X formation to prevent movement.
     */
    public void setX() {
        m_frontLeft.setDesiredState(new SwerveModuleState(0, Rotation2d.fromDegrees(45)));
        m_frontRight.setDesiredState(new SwerveModuleState(0, Rotation2d.fromDegrees(-45)));
        m_rearLeft.setDesiredState(new SwerveModuleState(0, Rotation2d.fromDegrees(-45)));
        m_rearRight.setDesiredState(new SwerveModuleState(0, Rotation2d.fromDegrees(45)));
    }

    /**
     * Sets the swerve ModuleStates.
     *
     * @param desiredStates The desired SwerveModule states.
     */
    public void setModuleStates(SwerveModuleState[] desiredStates) {
        SwerveDriveKinematics.desaturateWheelSpeeds(
            desiredStates, DrivetrainConstants.kMaxSpeedMetersPerSecond);
        m_frontLeft.setDesiredState(desiredStates[0]);
        m_frontRight.setDesiredState(desiredStates[1]);
        m_rearLeft.setDesiredState(desiredStates[2]);
        m_rearRight.setDesiredState(desiredStates[3]);
    }

    /** Resets the drive encoders to currently read a position of 0. */
    public void resetEncoders() {
        m_frontLeft.resetEncoders();
        m_rearLeft.resetEncoders();
        m_frontRight.resetEncoders();
        m_rearRight.resetEncoders();
    }

    /** Zeroes the heading of the robot. 
     * @return */
    public Command zeroHeading() {
        return run(() -> {
            gyro.reset();
            m_estimator.resetRotation(new Rotation2d(0));
        });
    }

    public double getGyroValue() {
        return gyro.getAngle() * (DrivetrainConstants.kGyroReversed ? -1 : 1);
    }

    /**
     * Returns the heading of the robot.
     *
     * @return the robot's heading in degrees, from -180 to 180
     */
    public double getHeading() {
        return Rotation2d.fromDegrees(getGyroValue()).getDegrees();
    }

    public TimeInterpolatableBuffer<Double> getGyroBuffer(){
        return gyroBuffer;
    }

    /**
     * Returns the turn rate of the robot.
     *
     * @return The turn rate of the robot, in degrees per second
     */
    public double getTurnRate() {
        return gyro.getRate() * (DrivetrainConstants.kGyroReversed ? -1.0 : 1.0);
    }

    public double driveDistance(){
        return m_frontLeft.getTotalDist();
    }

    public double getVelocity(){
        return m_frontLeft.getState().speedMetersPerSecond;
    }

    public Command resetToVision(){
        return run(() -> {
            m_estimator.resetPose(orangePose2d);
        });
    }
}
